import numpy as np
import pytest
from sklearn.datasets import load_breast_cancer

import pca_impute

np.random.seed(0)


Data = tuple[np.ndarray, np.ndarray]


def add_nans(X: np.ndarray, p: float = 0.2) -> np.ndarray:
    isnan = np.random.rand(*X.shape) < p
    X = np.where(isnan, np.nan, X)
    return X


@pytest.fixture(scope="session")
def data() -> Data:
    X0 = load_breast_cancer()["data"]
    X = add_nans(X0)
    return X0, X


def test_impute(data: Data) -> None:
    X0, X = data
    X_mean = pca_impute._impute._impute_by_mean(X)
    X_pca = pca_impute.impute(X, n_components=3)
    SST = np.sum((X0 - X_mean) ** 2)
    SSR = np.sum((X0 - X_pca) ** 2)
    R2 = 1 - SSR / SST
    # We expect pca should be much better than mean imputation.
    assert R2 > 0.8


def test_impute_auto_n_components(data: Data) -> None:
    X0, X = data
    X_pca_1 = pca_impute.impute(X, n_components=1)
    X_pca_auto = pca_impute.impute(X, n_components="auto")
    SSR_pca_1 = np.sum((X0 - X_pca_1) ** 2)
    SSR_pca_auto = np.sum((X0 - X_pca_auto) ** 2)
    assert SSR_pca_auto < SSR_pca_1



from typing import List, Union

import numpy as np
import scipy


def impute(X: np.ndarray, n_components: Union[int, str] = "auto", n_iter: int = 5) -> np.ndarray:
    """Impute the missing values by iteratively refining the guess by PCA reconstruction.

    Args:
        X: 2-d array with missing values represented by np.nan.
        n_components: The number of components of PCA model. When "auto",
            it is automatically chosen `pca_impute._impute._select_n_components`.
        n_iter: Number of refinement of guessed missing values.

    Returns:
        2-d array which is X with its missing value imputed.
    """
    if n_components == "auto":
        n_components = _select_n_components(X, n_iter)
    assert isinstance(n_components, (int, np.integer))

    isnan = np.isnan(X)
    mean = np.nanmean(X, axis=0)
    std = np.nanstd(X, axis=0)
    X = (X - mean[None, :]) / std
    X[isnan] = 0
    for i in range(n_iter):
        _, V = scipy.linalg.eigh(
            X.T @ X,
            subset_by_index=[X.shape[1] - n_components, X.shape[1] - 1],
        )
        X_approx = X @ V @ V.T
        X = np.where(isnan, X_approx, X)
    return X * std + mean


def _impute_by_mean(X: np.ndarray) -> np.ndarray:
    return np.where(np.isnan(X), np.nanmean(X, axis=0, keepdims=True), X)


def _get_candidate_n_components(n_components: int) -> List[int]:
    max_n_conponents = min(100, int(n_components / 2))
    return list(set(np.geomspace(1, max_n_conponents, 10, dtype=int)))


def _select_n_components(X: np.ndarray, n_iter: int) -> int:
    """Evaluate the quality of imputation by imputing artificially dropped values."""
    is_nan = np.isnan(X)
    p_missing = np.mean(is_nan, axis=0)
    should_drop = np.logical_and(np.random.rand(*X.shape) < p_missing[None, :], ~is_nan)
    X_dropped = X.copy()
    X_dropped[should_drop] = np.nan

    X_baseline = _impute_by_mean(X_dropped)
    candidate_n_components = _get_candidate_n_components(X.shape[1])
    scores = []
    for c in candidate_n_components:
        X_filled = impute(X_dropped, c, n_iter)
        SST = np.nansum((X_baseline - X) ** 2)
        SSR = np.nansum((X_filled - X) ** 2)
        R2 = 1 - SSR / SST
        scores.append(R2)
    return int(candidate_n_components[np.argmax(scores)])
